/*
 * @Author: luzc
 * @Date: 2019-10-01 17:01:44
 * @LastEditTime: 2019-10-09 09:04:56
 * @LastEditors: lzc
 * @Description: 开发环境服务调用
 */

process.env.NODE_ENV = 'development';

const chalk = require('chalk');
const clearConsole = require('react-dev-utils/clearConsole');
const formatWebpackMessages = require('react-dev-utils/formatWebpackMessages');
const webpack = require('webpack');
const WebpackDevServer = require('webpack-dev-server');
const openBrowser = require('react-dev-utils/openBrowser');
const prepareUrLs = require('../utils/prepareUrl');
const log = require('../utils/log');
const checkDepsInstalled = require('../utils/checkDepsInstalled');

module.exports = async function(context, subprocess) {
  const { applyHook, commandArgs, rootDir, webpackConfig } = context;
  log.verbose('dev cliOptions', commandArgs);

  await applyHook('beforeDev');

  const send = data => {
    if (subprocess && typeof subprocess.send === 'function') {
      subprocess.send(data);
    }
  };

  const installedDeps = checkDepsInstalled(rootDir);
  if (!installedDeps) {
    return Promise.reject(new Error('项目依赖未安装，请先安装依赖。'));
  }

  const HOST = commandArgs.host || '0.0.0.0';
  const PORT = commandArgs.port || 8000;

  const isInteractive = false; // process.stdout.isTTY;
  const urls = prepareUrLs('http', HOST, PORT);

  if (commandArgs.disabledReload) {
    log.warn('关闭了热更新（hot-reload）功能');
  }

  let isFirstCompile = true;

  if (commandArgs.disabledMock) {
    log.warn('关闭了 mock 功能');
  } else {
    // devMock 在 devServer 之前启动
    const originalDevServeBefore = webpackConfig.devServer.before;
    webpackConfig.devServer.before = function(app, server) {
      if (typeof originalDevServeBefore === 'function') {
        originalDevServeBefore(app, server);
      }
    };
  }

  const compiler = webpack(webpackConfig);
  const devServer = new WebpackDevServer(compiler, webpackConfig.devServer);

  compiler.hooks.done.tap('done', stats => {
    if (isInteractive) {
      clearConsole();
    }
    if (isFirstCompile) {
      send({
        action: 'update_project',
        message: 'server_finished',
        data: {
          statusDev: 'working',
          serverUrl: urls.localUrlForTerminal,
        },
      });

      isFirstCompile = false;
      openBrowser(urls.localUrlForBrowser);
    }

    applyHook('afterDev', stats);

    /*console.log(
      stats.toString({
        colors: true,
        chunks: false,
        assets: true,
        children: false,
        modules: false,
      })
    );*/
    console.log(
      [
        `\n`,
        `Local:   ${chalk.yellow(urls.localUrlForTerminal)}`,
        `Network: ${chalk.yellow(urls.lanUrlForTerminal)}`,
      ].join('\n')
    );

    const json = stats.toJson({}, true);
    const messages = formatWebpackMessages(json);
    const isSuccessful = !messages.errors.length && !messages.warnings.length;

    if (isSuccessful) {
      clearConsole();
      if (stats.stats) {
        log.info('编译成功！');
      } else {
        log.info(`编译成功，耗时 ${(json.time / 1000).toFixed(1)} 秒`);
      }

      // 服务启动完成切没有任务错误与警告
      send({
        action: 'update_project',
        message: 'compiler_success',
        data: {
          statusCompile: 'success',
          serverUrl: urls.lanUrlForBrowser || urls.localUrlForBrowser,
        },
      });
    } else {
      if (messages.errors.length) {
        if (messages.errors.length > 1) {
          messages.errors.length = 1;
        }
        console.log(messages.errors.join('\n\n'));
      } else if (messages.warnings.length) {
        log.warn('编译警告.');
        console.log();
        messages.warnings.forEach(message => {
          console.log(message);
          console.log();
        });
        console.log();
      }

      // 服务启动完成切没有任务错误与警告
      send({
        action: 'update_project',
        message: 'compiler_failed',
        data: {
          statusCompile: 'failed',
          serverUrl: urls.lanUrlForBrowser || urls.localUrlForBrowser,
        },
      });
    }
  });

  compiler.hooks.invalid.tap('invalid', () => {
    if (isInteractive) {
      clearConsole();
    }
    console.log('Compiling...');
    send({
      action: 'update_project',
      message: 'compiler_compiling',
      data: {
        statusCompile: 'compiling',
      },
    });
  });

  devServer.use((req, res, next) => {
    next();
  });

  devServer.listen(PORT, HOST, err => {
    if (err) {
      send({
        action: 'update_project',
        message: 'server_failed',
        data: {
          statusDev: 'failed',
        },
      });
      console.error(err);
      process.exit(500);
    } else {
      applyHook('afterDevServer', devServer);
      send({
        action: 'update_project',
        message: 'server_success',
        data: {
          statusDev: 'working',
          serverUrl: urls.lanUrlForBrowser || urls.localUrlForBrowser,
        },
      });
    }
  });
};
