/*
 * @Author: luzc
 * @Date:  2019/10/02
 * @Description:
 */

const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const ESLintFormatter = require('eslint-friendly-formatter');
const deepClone = require('lodash/cloneDeep');
const postcssConfig = require('./postcssConfig');
const getBabelConfig = require('./getBabelConfig');

const TYPESCRIPT_LOADER = require.resolve('ts-loader');
const BABEL_LOADER = require.resolve('babel-loader');
const ESLINT_LOADER = require.resolve('eslint-loader');
const CSS_LOADER = require.resolve('css-loader');
const LESS_LOADER = require.resolve('less-loader');
const POSTCSS_LOADER = require.resolve('postcss-loader');
// const SASS_LOADER = require.resolve('sass-loader');
const CSS_HOT_LOADER = require.resolve('css-hot-loader');
const URL_LOADER = require.resolve('url-loader');

const EXCLUDE_REGX = /node_modules/;
const URL_LOADER_LIMIT = 8192;

const defaultAssetsLoaderOpts = {
  name: 'assets/[hash].[ext]',
  limit: URL_LOADER_LIMIT
};

const cssLoaderOpts = {
  sourceMap: true
};

const cssModuleLoaderOpts = {
  ...cssLoaderOpts,
  modules: {
    mode: 'local',
    localIdentName: '[folder]--[local]--[hash:base64:7]'
  }
};

module.exports = (chainConfig, mode = 'development', zenConfig = {}) => {
  const babelConfig = getBabelConfig(zenConfig);

  function setExtraCSSLoader(lang, loaders) {
    const moduleTestReg = new RegExp(`\\.module\\.${lang}$`);
    const cssTestReg = new RegExp(`\\.${lang}$`);

    for (let i = 0; i < 2; i++) {
      const isModule = !!i;
      const ruleKey = `${lang}${isModule ? '-module' : ''}`;

      let rule = chainConfig.module.rule(ruleKey);

      if (isModule === true) {
        rule = rule.test(moduleTestReg);
      } else {
        rule = rule.test(cssTestReg);
        rule.exclude.add(moduleTestReg);
      }

      if (mode !== 'production') {
        rule.use('css-hot-loader').loader(CSS_HOT_LOADER);
      }

      rule
        .use('MiniCssExtractPlugin.loader')
        .loader(MiniCssExtractPlugin.loader);

      rule
        .use('css-loader')
        .loader(CSS_LOADER)
        .options(isModule ? cssModuleLoaderOpts : cssLoaderOpts);

      rule
        .use('postcss-loader')
        .loader(POSTCSS_LOADER)
        .options(Object.assign({ sourceMap: true }, postcssConfig));

      if (loaders && loaders.length > 0) {
        loaders.forEach(loader => {
          const [loaderName, loaderPath, loaderOptions] = loader;

          rule
            .use(loaderName)
            .loader(loaderPath)
            .options(Object.assign({ sourceMap: true }, loaderOptions));
        });
      }
    }
  }

  function setAssetsLoader(type, test, loaderOpts) {
    const rule = chainConfig.module.rule(type).test(test);

    rule
      .use(type)
      .loader(URL_LOADER)
      // keep defaultAssetsLoaderOpts immutable
      .options(Object.assign({}, defaultAssetsLoaderOpts, loaderOpts));
  }

  // css loader
  setExtraCSSLoader('css');
  // setExtraCSSLoader('scss', [['sass-loader', SASS_LOADER, {}]]);
  setExtraCSSLoader('less', [
    ['less-loader', LESS_LOADER, { sourceMap: true, javascriptEnabled: true }]
  ]);

  // assets loader
  setAssetsLoader('woff2', /\.woff2?$/, { mimetype: 'application/font-woff' });
  setAssetsLoader('ttf', /\.ttf$/, { mimetype: 'application/octet-stream' });
  setAssetsLoader('eot', /\.eot$/, {
    mimetype: 'application/vnd.ms-fontobject'
  });
  setAssetsLoader('svg', /\.svg$/, { mimetype: 'image/svg+xml' });
  setAssetsLoader('img', /\.(png|jpg|jpeg|gif)$/i);

  // 设置eslint
  if (mode === 'development' && zenConfig.eslint) {
    chainConfig.module
      .rule('eslint')
      .test(/\.(jsx|js)?$/)
      .exclude.add(EXCLUDE_REGX)
      .end()
      .enforce('pre')
      .use('eslint')
      .loader(ESLINT_LOADER)
      .options({
        cacheDirectory: true,
        formatter: ESLintFormatter
      });
  }

  // 开发环境忽略 node_modules文件夹
  if (mode === 'development') {
    // jsx loader
    chainConfig.module
      .rule('jsx')
      .test(/\.(jsx|js)?$/)
      .exclude.add(EXCLUDE_REGX)
      .end()
      .use('babel-loader')
      .loader(BABEL_LOADER)
      .options(
        Object.assign({}, deepClone(babelConfig), { cacheDirectory: true })
      );

    // tsx loader
    chainConfig.module
      .rule('tsx')
      .test(/\.(tsx|ts)?$/)
      .exclude.add(EXCLUDE_REGX)
      .end()
      .use('babel-loader')
      .loader(BABEL_LOADER)
      .options(
        Object.assign({}, deepClone(babelConfig), { cacheDirectory: true })
      )
      .end()
      .use('ts-loader')
      .loader(TYPESCRIPT_LOADER)
      .options({ transpileOnly: true });
  }
};
