/*
 * @Author: luzc
 * @Date: 2019-10-01 15:06:34
 * @LastEditTime: 2019-10-09 09:05:25
 * @LastEditors: lzc
 * @Description:
 */
const fs = require('fs');
const path = require('path');
const fse = require('fs-extra');
const assert = require('assert');
const deepmerge = require('deepmerge');

const log = require('../utils/log');
const getDefaultWebpackConfig = require('../config/getDefaultWebpackConfig');
const processEntry = require('../config/processEntry');
const PluginAPI = require('./Plugin');
const defaultConfig = require('../config/default.config');

module.exports = class Context {
  constructor({ command = '', rootDir = process.cwd(), args = {} }) {
    this.command = command;
    this.commandArgs = args;
    this.rootDir = rootDir;
    this.pkg = this.getPkgData(this.rootDir);

    this.userConfig = this.getUserConfig();
    this.plugins = this.getPlugins();

    this.chainWebpackFns = [];
    this.eventHooks = {};

    this.reRun = this.reRun.bind(this);
    this.applyHook = this.applyHook.bind(this);
  }

  /**
   * 获取package.json
   */
  getPkgData() {
    const pkg = {};
    try {
      Object.assign(
        pkg,
        JSON.parse(fs.readFileSync(path.resolve(this.rootDir, 'package.json'), 'utf-8'))
      );
    } catch (err) {
      log.error('获取项目 package.json 失败, 检查是否有语法错误. ');
      throw err;
    }
    return pkg;
  }

  /**
   * 获取zen.config.js
   */
  getUserConfig() {
    const { config } = this.commandArgs;
    let zenConfigPath = '';
    if (config) {
      zenConfigPath = path.isAbsolute(config) ? config : path.resolve(this.rootDir, config);
    } else {
      zenConfigPath = path.resolve(this.rootDir, 'zend.config.js');
    }

    let userConfig = {};
    if (fse.existsSync(zenConfigPath)) {
      try {
        userConfig = require(zenConfigPath);
      } catch (err) {
        log.error(`加载配置文件失败 ${zenConfigPath}, 请检查配置是否正确`);
        console.error(err);
        process.exit(1);
      }
    }
    return deepmerge(defaultConfig, userConfig);
  }

  /**
   * 获取插件
   * @returns {any[]}
   */
  getPlugins() {
    const builtInPlugins = [
      '../plugins/userConfig',
      '../plugins/cliOptions',
      '../plugins/devWatcher',
    ];
    return builtInPlugins.map(pluginPath => require(pluginPath))
      .concat(this.userConfig.plugins || []);
  }

  /**
   * 运行插件
   * @returns {Promise<void>}
   */
  async runPlugins() {
    for (let i = 0; i < this.plugins.length; i++) {
      let pluginInfo = this.plugins[i];
      if (!Array.isArray(pluginInfo)) {
        pluginInfo = [pluginInfo];
      }
      const [plugin, options] = pluginInfo;
      const pluginName = typeof plugin === 'string' ? plugin : '';
      try {
        const pluginFunc = typeof plugin === 'string'
          ? require(require.resolve(plugin, { paths: [this.rootDir] }))
          : plugin;
        assert(typeof pluginFunc === 'function', 'plugin must export a function');
        await pluginFunc(new PluginAPI(this, pluginName), options);
      } catch (err) {
        log.error(`Fail to load plugin ${pluginName}`);
        console.error(err);
        process.exit(1);
      }
    }
  }

  /**
   * 运行webpack钩子
   * @param key
   * @param opts
   * @returns {Promise<void>}
   */
  async applyHook(key, opts = {}) {
    const hooks = this.eventHooks[key] || [];
    const results = [];
    hooks.forEach((fn) => {
      try {
        results.push(fn(opts));
      } catch (e) {
        log.error(e);
        log.error(`Fail to execute hook ${key}`);
      }
    });
    await Promise.all(results);
  }

  /**
   * 获取配置
   * @returns {*}
   */
  getWebpackConfig() {
    const config = getDefaultWebpackConfig(this.command, this.userConfig);

    this.chainWebpackFns.forEach(({ fn, pluginName }) => {
      try {
        fn(config);
      } catch (err) {
        log.error(`Fail to exec plugin chainWebpack ${pluginName}`);
        throw (err);
      }
    });

    if (this.userConfig.chainWebpack) {
      try {
        this.userConfig.chainWebpack(config, {
          command: this.command,
          commandArgs: this.commandArgs,
        });
      } catch (err) {
        log.error('Fail to exec userConfig.chainWebpack');
        throw (err);
      }
    }

    processEntry(config, {
      hotDev: this.command === 'dev' && !this.commandArgs.disabledReload,
      polyfill: this.userConfig.injectBabel === 'polyfill',
    });

    return config.toConfig();
  }

  /**
   * 重启
   * @returns {Promise<void>}
   */
  async reRun() {
    this.chainWebpackFns = [];
    this.eventHooks = {};
    this.run();
  }

  /**
   * 运行
   * @returns {Promise<void>}
   */
  async run() {
    await this.runPlugins();
    this.webpackConfig = this.getWebpackConfig();
    const command = require(`../commands/${this.command}`);
    await command(this);
  }
};
